<?php

namespace Drp\DeepReplicate;

use Illuminate\Database\Eloquent\Collection;
use Illuminate\Database\Eloquent\Model;

class DeepReplicator
{
    /**
     * A cache of objects created by the replicator to avoid duping things twice.
     * @var array
     */
    protected $cache = [];

    /**
     * Recursively call replicate on a models relations to clone the whole tree
     * @param Model $model The model to replicate.
     * @param array $except Attributes to exclude.
     * @return Model The new model with the replicated relations.
     */
    public function deepReplicate(Model $model, array $except = null)
    {
        // clear any loaded relations
        $model->setRelations([]);

        // make a new model
        $newModel = $this->replicate($model, $except);

        // TODO possibly remove this, it's a bit specific to multievent
        if ($newModel->isFillable('original_id')) {
            $newModel->original_id = $model->id;
        }

        $newModel->save();

        // Load the relationships that we want to clone so we can loop over them.
        $relations = $this->getRelationsToReplicate($model);

        $model->load($relations);

        foreach ($model->getRelations() as $relationName => $relationItems) {
            $relationItems = $this->ensureCollection($relationItems);

            // Go through each item on the relation, create a copy and save it against the new model
            foreach ($relationItems as $relationItem) {
                $relationItemCopy = $this->deepReplicate($relationItem);
                $relationQuery = $newModel->{$relationName}();

                if (method_exists($relationQuery, 'associate')) {
                    $relationQuery->associate($relationItemCopy)->save();
                    continue;
                }

                $newModel->{$relationName}()->save($relationItemCopy);
            }
        }

        return $newModel;
    }

    /**
     * Return the relations to replicate when fully replicating the model.
     * @param Model $model The model to check.
     * @return array
     */
    protected function getRelationsToReplicate(Model $model)
    {
        if (property_exists($model, 'replicableRelations')) {
            return $model->replicableRelations;
        }

        return [];
    }

    /**
     * Make sure the passed variable is always a collection.
     * @param mixed $item The item to consider.
     * @return \Illuminate\Support\Collection
     */
    protected function ensureCollection($item)
    {
        if (is_null($item)) {
            return collect();
        }

        if ($item instanceof Collection) {
            return $item;
        }

        if ($item instanceof Model) {
            return collect([$item]);
        }

        return collect($item);
    }

    /**
     * Replicate a model, but only if it's not already been done.
     * We use a cache to avoid situations where many to many relations might duplicate the same object
     * multiple times as it loops through. By keying on the original model id, and storing a reference to the clone, we
     * can update the relationship instead of creating a new object.
     * @param Model $model The model to replicate.
     * @param array|null $except The attributes not to replicate.
     * @return Model
     */
    protected function replicate(Model $model, array $except = null)
    {
        $key = $model->getTable() . "." . $model->getKey();

        if (array_has($this->cache, $key)) {
            return array_get($this->cache, $key);
        }
        $newModel = $model->replicate($except);

        array_set($this->cache, $key, $newModel);

        return $newModel;
    }
}
